; -*- Scheme -*-

(load "fold.scm")

;; bindings := (<binding>...)
;; binding  := (<tag> <bandler-pair>...)
;;           | (*default* . <post-handler>)
;;           | (*text* . <text-handler>)
;; tag      := <symbol>
;; handler-pair := (pre-layout . <pre-layout-handler>)
;;           | (post . <post-handler>)
;;           | (bindings . <bindings>)
;;           | (pre . <pre-handler>)
;;           | (macro . <macro-handler>)
;;
;; pre-layout-handler is a function of three arguments:
;;   kids: the kids of the current node, before traversal
;;   params: the params of the current node
;;   layout: the layout coming into this node
;; pre-layout-handler is expected to use this information to return a
;; layout to pass to the kids. The default implementation returns the
;; layout given in the arguments.
;;
;; post-handler is a function of five arguments:
;;   tag: the current tag being processed
;;   params: the params of the current node
;;   layout: the layout coming into the current node, before any kids
;;     were processed
;;   klayout: the layout after processing all of the children
;;   kids: the already-processed child nodes
;; post-handler should return two values, the layout to pass to the next
;; node and the final tree.
;;
;; text-handler is a function of three arguments:
;;   text: the string
;;   params: the current params
;;   layout: the current layout
;; text-handler should return two values, the layout to pass to the next
;; node and the value to which the string should transform.

(define (assq-ref alist key default)
  (cond ((assq key alist) => cdr)
        (else default)))

(define (fold-layout tree bindings params layout)
  (define (err . args)
    (error "no binding available" args))
  (define (fdown tree bindings pcont params layout ret)
    (define (fdown-helper new-bindings new-layout cont)
      (let ((cont-with-tag (lambda args
                             (apply cont (car tree) args)))
            (bindings (if new-bindings
                          (append new-bindings bindings)
                          bindings)))
        (cond
         ((null? (cdr tree))
          (values
           '() bindings cont-with-tag (cons '() params) new-layout '()))
         ((and (pair? (cadr tree)) (eq? (caadr tree) '@))
          (let ((params (cons (cdadr tree) params)))
            (values
             (cddr tree) bindings cont-with-tag params new-layout '())))
         (else
          (values
           (cdr tree) bindings cont-with-tag (cons '() params) new-layout '())))))
    (define (no-bindings)
      (fdown-helper #f layout (assq-ref bindings '*default* err)))
    (define (macro macro-handler)
      (fdown (apply macro-handler tree)
             bindings pcont params layout ret))
    (define (pre pre-handler)
      (values '() bindings 
              (lambda (params layout old-layout kids)
                (values layout (reverse kids)))
              params layout (apply pre-handler tree)))
    (define (have-bindings tag-bindings)
      (fdown-helper
       (assq-ref tag-bindings 'bindings #f)
       ((assq-ref tag-bindings 'pre-layout
                  (lambda (tag params layout)
                    layout))
        tree params layout)
       (assq-ref tag-bindings 'post
                 (assq-ref bindings '*default* err))))
    (let ((tag-bindings (assq-ref bindings (car tree) #f)))
      (cond
       ((not tag-bindings) (no-bindings))
       ((assq-ref tag-bindings 'macro #f) => macro)
       ((assq-ref tag-bindings 'pre #f) => pre)
       (else (have-bindings tag-bindings)))))
  (define (fup tree bindings cont params layout ret
               kbindings kcont kparams klayout kret)
    (call-with-values
        (lambda ()
          (kcont kparams layout klayout (reverse kret)))
      (lambda (klayout kret)
        (values bindings cont params klayout (cons kret ret)))))
  (define (fhere tree bindings cont params layout ret)
    (call-with-values
        (lambda ()
          ((assq-ref bindings '*text* err) tree params layout))
      (lambda (tlayout tret)
        (values bindings cont params tlayout (cons tret ret)))))
  (call-with-values
      (lambda ()
        (foldts*-values
         fdown fup fhere tree bindings #f (cons params '()) layout '()))
    (lambda (bindings cont params layout ret)
      (values (car ret) layout))))
