; -*- Scheme -*-

(load "fold.scm")

(define-macro (assert-equal X Y)
  (let ((x (gensym)) (y (gensym)))
    `(let ((,x ,X) (,y ,Y))
       (or (equal? ,x ,y)
           (begin
             (newline)
             (write ',X)
             (display " = ")
             (write ,x)
             (newline)
             (newline)
             (write ',Y)
             (display " = ")
             (write ,y)
             (newline)
             (newline)
             (error "not equal"))))))

(define-macro (assert X)
  (let ((x (gensym)))
    `(let ((,x ,X))
       (or ,x
           (error "not true:" ',X "=" ,x)))))

(define test-doc
  '(presentation
    (@ (width 1024)
       (height 768)
       (title-style "font-family:Georgia")
       (title-height 72)
       (title-baseline-y 96)
       (title-x 48)
       (text-height 64)
       (text-style "font-family:Georgia")
       (text-upper-left-x 96)
       (text-upper-left-y 216))
    (slide
     (@ (title "Declarative interface"))
     (p "The declarative interface"
        "lets you be more concise"
        "when making the slides."))
    (slide
     (@ (title "Still cumbersome"))
     (p "Parentheses are still"
        "cumbersome."))))

(assert (atom? 'foo))
(assert (atom? '()))
(assert (not (atom? '(1 2 3))))

(define (id x) x)
(assert-equal (foldt id id test-doc) test-doc)

(assert-equal (fold cons '() test-doc)
              (reverse test-doc))

(assert-equal (foldts (lambda (seed tree) '())
                      (lambda (seed kid-seed tree)
                        (cons (reverse kid-seed) seed))
                      (lambda (seed tree)
                        (cons tree seed))
                      '()
                      test-doc)
              (cons test-doc '()))

(assert-equal (foldts* (lambda (seed tree) (values '() tree))
                       (lambda (seed kid-seed tree)
                         (cons (reverse kid-seed) seed))
                       (lambda (seed tree)
                         (cons tree seed))
                       '()
                       test-doc)
              (cons test-doc '()))

(assert-equal (fold-values cons test-doc '())
              (fold cons '() test-doc))

(assert-equal (foldts*-values
               (lambda (tree seed) (values tree '()))
               (lambda (tree seed kid-seed)
                 (cons (reverse kid-seed) seed))
               (lambda (tree seed)
                 (cons tree seed))
               test-doc
               '())
              (foldts* (lambda (seed tree) (values '() tree))
                       (lambda (seed kid-seed tree)
                         (cons (reverse kid-seed) seed))
                       (lambda (seed tree)
                         (cons tree seed))
                       '()
                       test-doc))

(define (replace pred val list)
  (reverse
   (fold
    (lambda (x xs)
      (cons (if (pred x) val x) xs))
    '()
    list)))

(define (car-eq? x what)
  (and (pair? x) (eq? (car x) what)))

;; avoid entering <slide>
(assert-equal (foldts*-values
               (lambda (tree seed)
                 (values (if (car-eq? tree 'slide) '() tree) '()))
               (lambda (tree seed kid-seed)
                 (cons (reverse kid-seed) seed))
               (lambda (tree seed)
                 (cons tree seed))
               test-doc
               '())
              (cons
               (replace (lambda (x) (car-eq? x 'slide))
                        '()
                        test-doc)
               '()))

(define (all-elts tree)
  (reverse!
   (foldts*-values
    (lambda (tree seed)
      (values tree seed))
    (lambda (tree seed kid-seed)
      kid-seed)
    (lambda (tree seed)
      (cons tree seed))
    tree
    '())))

(define (len tree)
  (foldts*-values
   (lambda (tree seed)
     (values tree seed))
    (lambda (tree seed kid-seed)
      kid-seed)
    (lambda (tree seed)
      (1+ seed))
    tree
    0))

(assert-equal (length (all-elts test-doc))
              (len test-doc))

(display "all tests pass") (newline)
