; -*- Scheme -*-

(load "pre-post-order.scm")

(define-macro (assert-equal X Y)
  (let ((x (gensym)) (y (gensym)))
    `(let ((,x ,X) (,y ,Y))
       (or (equal? ,x ,y)
           (begin
             (newline)
             (write ',X)
             (display " = ")
             (write ,x)
             (newline)
             (newline)
             (write ',Y)
             (display " = ")
             (write ,y)
             (newline)
             (newline)
             (error "not equal"))))))

(define hello-doc
  '(slides
    (slide
     (title "Foo")
     (para "Hello" (br) "world"))))

(define slides-to-html
  `((slides . ,(lambda (tag . kids)
                 `(html (body ,@kids))))
    (slide  . ,(lambda (tag . kids)
                 `(div (@ (class "slide")) ,@kids)))
    (title  . ,(lambda (tag . kids)
                 `(h1 ,@kids)))
    (para   . ,(lambda (tag . kids)
                 `(p ,@kids)))
    (br     . ,(lambda (tag . kids)
                 `(br)))
    (*text* . ,(lambda (_ txt) txt))))

(define expected-html
  '(html
    (body
     (div (@ (class "slide"))
          (h1 "Foo")
          (p "Hello"
             (br)
             "world")))))

(assert-equal (post-order/foldt slides-to-html hello-doc)
              expected-html)
(assert-equal (post-order slides-to-html hello-doc)
              expected-html)

(define slides-to-html-nested
  `((slides
     ((slide
       ((title
         ((*text* . ,(lambda (_ txt) txt)))
         . ,(lambda (tag . kids)
              `(h1 ,@kids)))
        (para
         ((*text* . ,(lambda (_ txt) txt))
          (br     . ,(lambda (tag . kids)
                       `(br))))
         . ,(lambda (tag . kids)
              `(p ,@kids))))
       . ,(lambda (tag . kids)
            `(div (@ (class "slide")) ,@kids))))
     . ,(lambda (tag . kids)
          `(html (body ,@kids))))))

(assert-equal (post-order slides-to-html-nested hello-doc)
              expected-html)

(assert-equal (pre-post-order slides-to-html-nested hello-doc)
              expected-html)

(assert-equal (pre-post-order
               `((foo *preorder* . ,(lambda args '(bar))))
               '(foo (unknown "invalid")))
              '(bar))

(assert-equal (pre-post-order
               `((foo *macro* . ,(lambda args '(bar)))
                 (bar *macro* . ,(lambda args '(baz)))
                 (baz . ,(lambda (tag) '(qux))))
               '(foo (unknown "invalid")))
              '(qux))

(display "all tests pass") (newline)
