(load "fold.scm")

(define (assq-ref alist key default)
  (cond ((assq key alist) => cdr)
        (else default)))

(define (post-order/foldt bindings tree)
  (let* ((err (lambda args (error "no binding available")))
         (default (assq-ref bindings '*default* err)))
    (define (fup kids)
      (apply (assq-ref bindings (car kids) default) kids))
    (define (fhere txt)
      ((assq-ref bindings '*text* err) '*text* txt))
    (foldt fup fhere tree)))

(define (post-order bindings tree)
  (define (err . args)
    (error "no binding available" args))
  (define (fdown tree bindings pcont ret)
    (let ((tail (assq-ref bindings (car tree) #f)))
      (cond
       ((not tail)
        (let ((default (assq-ref bindings '*default* err)))
          (values tree bindings default '())))
       ((pair? tail)
        (let ((new-bindings (append (car tail) bindings))
              (cont (cdr tail)))
          (values tree new-bindings cont '())))
       (else
        (values tree bindings tail '())))))
  (define (fup tree bindings cont ret kbindings kcont kret)
    (values bindings cont (cons (apply kcont (reverse kret)) ret)))
  (define (fhere tree bindings cont ret)
    (define (tcont x)
      (if (symbol? x)
          x
          ((or (assq-ref bindings '*text* #f)
               (assq-ref bindings '*default* err))
           '*text* x)))
    (values bindings cont (cons (tcont tree) ret)))
  (call-with-values
      (lambda ()
        (foldts*-values fdown fup fhere tree bindings #f '()))
    (lambda (bindings cont ret)
      (car ret))))

(define (pre-post-order bindings tree)
  (define (err . args)
    (error "no binding available" args))
  (define (fdown tree bindings pcont ret)
    (let ((tail (assq-ref bindings (car tree) #f)))
      (cond
       ((not tail)
        (let ((default (assq-ref bindings '*default* err)))
          (values tree bindings default '())))
       ((pair? tail)
        (let ((cont (cdr tail)))
          (case (car tail)
           ((*preorder*)
            (values '() bindings 
                    (lambda x (reverse x)) (apply cont tree)))
           ((*macro*)
            (fdown (apply cont tree) bindings pcont ret))
           (else
            (let ((new-bindings (append (car tail) bindings)))
              (values tree new-bindings cont '()))))))
       (else
        (values tree bindings tail '())))))
  (define (fup tree bindings cont ret kbindings kcont kret)
    (values bindings cont (cons (apply kcont (reverse kret)) ret)))
  (define (fhere tree bindings cont ret)
    (define (tcont x)
      (if (symbol? x)
          x
          ((or (assq-ref bindings '*text* #f)
               (assq-ref bindings '*default* err))
           '*text* x)))
    (values bindings cont (cons (tcont tree) ret)))
  (call-with-values
      (lambda ()
        (foldts*-values fdown fup fhere tree bindings #f '()))
    (lambda (bindings cont ret)
      (car ret))))
